/*
 * Choice JS
*/

class Builder {

    //Constructor
    constructor(target, userOptions){
        this.target = target;
        //Build our default options
        this.options = {
            ...this.buildDefaultOptions(userOptions),
            ...userOptions
        };
    }
    
    
    /* ------------------------------------------------------------------------------------------------------------------------------------------------
    |  Methods, other functions and methods
    ------------------------------------------------------------------------------------------------------------------------------------------------ */
    

    /**
     * Create a new choice dropdown
     * 
     * @return {Promise}
    */
    async create(){
        //Build it
        let choice = await this.build();
        //Toast bindings
        await this.bindings(choice);
        //Configure the select based on the options
        await this.configureOptions(choice);
        //Hide the current
        this.target.setAttribute('data-choicejs-id', `#${choice.id}`);
        //Deploy the select
        this.target.parentNode.insertBefore(choice, this.target.nextSibling);
        //Preselect any selected options
        await this.selectPreselectedOptions(choice);
        //Done
        return choice;
    }


    /**
     * Configure options for the choice
     * 
     * @param {HTMLElement} choice - The select we're configuring options for
     * @return {Promise}
    */
    async configureOptions(choice){
        
        //Width and height
        choice.style.width = (this.options.width == 'inherit') ? (this.target.offsetWidth + 'px') : this.options.width,
        choice.style.minHeight = (this.options.height == 'inherit') ? (this.target.offsetHeight + 'px') : this.options.height;
        
        //Class and id
        if(this.options.id != '') choice.setAttribute('id', this.options.id)
        if(this.options.class != '') choice.classList.add(this.options.class);
    
        //Placeholder
        if(this.options.placeholder) choice.querySelector('.choicejs-select').innerHTML = this.options.placeholder;
        
        //Reset the selected index (incase set by placeholder)
        this.target.selectedIndex = -1;
        
        //Return
        return;
    }
    
    
    /**
     * Build the default options
     * 
     * @param {Object} userOptions The user options for any overwrites
     * @return {Object} An object of options
    */
    buildDefaultOptions(userOptions){
        let options = {
            width:'inherit',
            height:'inherit',
            placeholder: (this.target.hasAttribute('placeholder') ? this.target.getAttribute('placeholder') : false),
            id: '',
            class: '',
            closeOnSelect: false,
            event: 'change'
        }
        //Return
        return options;
    }

    
    /**
     * Add option
     *
     * @param {HTMLElement} option - The option we're selecting
     * @param {HTMLElement} choice - The choice instance
     * @returns {Promise}
    */
    async addOption(option, choice){
        //Get elems
        let value = option.getAttribute('value'),
            select = choice.querySelector('.choicejs-select'),
            multiple = (this.target.hasAttribute('multiple') && this.target.getAttribute('multiple') == 'true');
        //Single select or multi?
        if(multiple){
            //Build and add a pill
            let pill = await this.buildOptionPill(option);
            await this.bindPillRemoveBtn(pill, choice);
            //If there is a placeholder, clear it first
            if(select.querySelectorAll('.choicejs-pill').length == 0) select.innerHTML = '';
            select.appendChild(pill);
        } else {
            //Deselect all other options
            for(let sibling of choice.querySelectorAll('li')){
                sibling.classList.remove('selected');
            }
            //Set the select value
            select.innerHTML = option.innerHTML;
            //Deselect all hidden options
            for(let hiddenOption of this.target.querySelectorAll('option')){
                hiddenOption.removeAttribute('selected');
            }
        }
        //Select the hidden option
        this.target.querySelector(`option[value="${value}"]`).selected = true;
        //Add the selected class
        option.classList.add('selected');
    }
    
    
    /**
     * Remove option
     *
     * @param {HTMLElement} option - The option we're selecting
     * @param {HTMLElement} choice - The choice instance
     * @returns {Promise}
    */
    async removeOption(option, choice){
        //Get elems
        let value = option.getAttribute('value'),
            select = choice.querySelector('.choicejs-select'),
            multiple = (this.target.hasAttribute('multiple') && this.target.getAttribute('multiple') == ('true' || 'multiple'));
        //Remove the hidden option
        this.target.querySelector(`option[value="${value}"]`).selected = false;
        this.target.querySelector(`option[value="${value}"]`).setAttribute('selected', false);
        //Remove the selected class on the option
        option.classList.remove('selected');
        //Single or multiple select?
        if(multiple){
            //Remove the pill
            let pill = select.querySelector(`.choicejs-pill[value="${value}"]`);
            if(pill) pill.remove();
            //If there are pills left, return
            if(select.querySelectorAll('.choicejs-pill').length > 0) return;
        }
        //Set the placeholder
        if(this.options.placeholder) select.innerHTML = this.options.placeholder;
        //Reset the select index
        this.target.selectedIndex = -1;
    }
    
    
    /**
     * Select pre-selected options. Loop through options that are already selected on page load and enable them
     *
     * @param {HTMLElement} choice - The instance of choice
     * @returns {Promise}
    */
    async selectPreselectedOptions(choice){
        //Loop the choice options
        for(let selected_option of choice.querySelectorAll('.choicejs-dropdown li.selected')){
            this.addOption(selected_option, choice);
        }
        return;
    }
      
    
    /* ------------------------------------------------------------------------------------------------------------------------------------------------
    |  Builders, methods that build HTML
    ------------------------------------------------------------------------------------------------------------------------------------------------ */
    
    
    /**
     * Build the choice
     * 
     * @return {Promise} HTMLElement The new choice
    */
    async build(){
        //Build the new Select HTML element and add it directly after the target
        let wrapper = document.createElement('div'),
            select = document.createElement('div'),
            search = document.createElement('input'),
            dropdown = document.createElement('div'),
            list = document.createElement('ul');
        //Set attributes
        wrapper.classList.add('choicejs');
        wrapper.setAttribute('id', 'choice-'+Math.random().toString(36).substr(2, 4));
        select.classList.add('choicejs-select');
        search.classList.add('choicejs-search');
        search.setAttribute('placeholder', 'Search');
        dropdown.classList.add('choicejs-dropdown');
        //Loop the targets options and build our list
        for(let option of this.target.querySelectorAll('option')){
            if(option.value != '' && option.value != undefined){
                let item = document.createElement('li');
                item.setAttribute('value', option.value);
                item.innerHTML = option.innerHTML;
                //If the option is selected, pre-select the item
                if(option.hasAttribute('selected')){
                    item.classList.add('selected');
                }
                list.appendChild(item);
            }
        }
        //Build the whole select and return it
        dropdown.appendChild(search);
        dropdown.appendChild(list);
        wrapper.appendChild(select);
        wrapper.appendChild(dropdown);
        //Return
        return wrapper;
    }
    
    
    /**
     * Build option pill
     *
     * @param {Node} option - The option selected
     * @returns {Promise} A node element
    */
    async buildOptionPill(option){
        let wrapper = document.createElement('div'),
            close = document.createElement('span'),
            value = option.getAttribute('value');
        //Attributes
        wrapper.classList.add('choicejs-pill');
        wrapper.setAttribute('value', value);
        wrapper.innerHTML = option.innerHTML;
        //Build it
        wrapper.appendChild(close);
        //Return
        return wrapper;
    }
    
    
    /* ------------------------------------------------------------------------------------------------------------------------------------------------
    |  Bindings, methods that are bindings to functionality
    ------------------------------------------------------------------------------------------------------------------------------------------------ */
    
    
    /**
     * Add all of the bindings for the choice
     * 
     * @param {HTMLElement} choice - The select we're building the bindings for
     * @return {Promise}
    */
    async bindings(choice){
        //Return
        return await Promise.all([
            this.bindActiveListener(choice),
            this.bindOptionSelect(choice),
            this.bindOptionFiltering(choice),
        ]);
    }
    
    
    /**
     * Bind active click
     *
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindActiveListener(choice){
        choice.addEventListener('click', () => {
            choice.classList.add('active');
        })
        document.addEventListener('click', function(event) {
            if (!choice.contains(event.target)) {
                choice.classList.remove('active');
            }
        });
        //Return
        return;
    }
    
    
    /**
     * Bind option select
     * 
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindOptionSelect(choice){
        //Loop and add
        for(let option of choice.querySelectorAll('ul li')){
            //Add the event listener
            option.addEventListener('click', async(e) => {
                //Stop Prop
                e.stopPropagation();
                e.stopImmediatePropagation();
                //Add or remove
                if(option.classList.contains('selected')){
                    await this.removeOption(option, choice);
                } else {
                    await this.addOption(option, choice);
                }
                //Close on select?
                if(this.options.closeOnSelect){
                    choice.classList.remove('active');
                }
                //What parent event to trigger?
                if(this.options.event !== false){
                    this.target.dispatchEvent(new Event(this.options.event));
                }
            });
        } 
        //Return
        return;
    }
    
    
    /**
     * Bind option filtering
     * 
     * @param {HTMLElement} choice - The choice we're binding
     * @return {Promise}
    */
    async bindOptionFiltering(choice){
        //When the choice search bar changes
        var searchTimeout = 0,
            searchBar = choice.querySelector('.choicejs-search'),
            options = choice.querySelectorAll('ul li');
        //When the postcode is changed, test it
        searchBar.addEventListener('input', () => {
            //Clear the above timeout
            clearTimeout(searchTimeout);
            //Set a new timeout
            searchTimeout = setTimeout(() =>{
                //Loop all options
                for(let option of options){
                    if(option.innerHTML.includes(searchBar.value)){
                        option.classList.remove('hide');
                    } else {
                        option.classList.add('hide');
                    }
                }
            }, 500);
        });
        //Return
        return;
    }
    
    
    /**
     * Bind pill remove btn
     *
     * @param {Node} pill - The pill
     * @param {HTMLElement} choice - The choice instance
     * @returns {Promise}
    */
    async bindPillRemoveBtn(pill, choice){
        //Params
        let value = pill.getAttribute('value'),
            option = choice.querySelector(`.choicejs-dropdown li[value="${value}"]`);
        //Add the remove event listener
        pill.querySelector('span').addEventListener('click', (e) => {
            //Run remove btn
            this.removeOption(option, choice);
        })
        //Return
        return;
    }
    

}





/**
 * Export the choice function to use
 *
 * @param {NodeElement} target - The target we're adding the choice to
 * @param {Object} options - The options
 * @returns {Promise}  - An instance of choice
 */
export const Choice = async (target, options) => {
    //Start a new Choice
    let newBuild = new Builder(target, options);
    //Run the build
    return await newBuild.create();
}